/* listomf.c (emx+gcc) */

/* This program lists OMF files (.obj files).

Copyright (c) 1993 Eberhard Mattes

This file is part of listomf.

listomf is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

listomf is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

As a special exception, listomf can be distributed with emx.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "defs.h"

#define MAX_REC_SIZE    1024

#define THEADR          0x80
#define COMENT          0x88
#define MODEND          0x8a
#define EXTDEF          0x8c
#define TYPDEF          0x8e
#define PUBDEF          0x90
#define LINNUM          0x94
#define LNAMES          0x96
#define SEGDEF          0x98
#define GRPDEF          0x9a
#define FIXUPP          0x9c
#define LEDATA          0xa0
#define LIDATA          0xa2
#define COMDEF          0xb0

#define REC32           0x01


#pragma pack(1)

struct omf_rec
{
  byte rec_type;
  word rec_len;
};

struct timestamp
{
  byte hours;
  byte minutes;
  byte seconds;
  byte hundredths;
  byte day;
  byte month;
  word year;
};

#pragma pack()

struct segment
{
  int name;
  int class;
  dword length;
};

struct group
{
  int name;
};

struct ext
{
  char *name;
};

enum debug
{
  debug_default,
  debug_cv,
  debug_hll
};


static int lname_count;
static int lname_alloc;
static char **lname_list;
static int segment_count;
static int segment_alloc;
static struct segment *segment_list;
static int group_count;
static int group_alloc;
static struct group *group_list;
static int ext_count;
static int ext_alloc;
static struct ext *ext_list;
static int pub_count;
static int type_index;
static int rec_type;
static int rec_len;
static int rec_idx;
static byte rec_buf[MAX_REC_SIZE+8];
static enum debug debug_info;
static dword line_count;
static dword names_length;
static int linnum_started;


static void out_of_mem (void)
{
  fputs ("Out of memory\n", stderr);
  exit (2);
}


static void *xrealloc (void *src, size_t n)
{
  void *p;

  p = realloc (src, n);
  if (p == NULL)
    out_of_mem ();
  return (p);
}


static char *xstrdup (const char *src)
{
  char *p;

  p = strdup (src);
  if (p == NULL)
    out_of_mem ();
  return (p);
}


static void show_record (const char *name)
{
  printf ("%s", name);
  if (rec_type & REC32)
    printf ("32");
}


static void get_mem (void *dst, int len)
{
  if (rec_idx + len > rec_len)
    {
      fprintf (stderr, "\nString beyond end of record\n");
      exit (2);
    }
  memcpy (dst, rec_buf + rec_idx, len);
  rec_idx += len;
}


static void get_string (byte *dst)
{
  int len;

  if (rec_idx >= rec_len)
    {
      fprintf (stderr, "\nString beyond end of record\n");
      exit (2);
    }
  len = rec_buf[rec_idx++];
  get_mem (dst, len);
  dst[len] = 0;
}


static int get_index (void)
{
  int result;

  if (rec_idx >= rec_len)
    {
      fprintf (stderr, "\nIndex beyond end of record\n");
      exit (2);
    }
  result = rec_buf[rec_idx++];
  if (result & 0x80)
    {
      if (rec_idx >= rec_len)
        {
          fprintf (stderr, "\nIndex beyond end of record\n");
          exit (2);
        }
      result = ((result & 0x7f) << 8) | rec_buf[rec_idx++];
    }
  return (result);
}


static word get_dword (void)
{
  dword result;

  if (rec_idx + 4 > rec_len)
    {
      fprintf (stderr, "\nDword beyond end of record\n");
      exit (2);
    }
  result = rec_buf[rec_idx++];
  result |= rec_buf[rec_idx++] << 8;
  result |= rec_buf[rec_idx++] << 16;
  result |= rec_buf[rec_idx++] << 24;
  return (result);
}


static word get_word (void)
{
  word result;

  if (rec_idx + 2 > rec_len)
    {
      fprintf (stderr, "\nWord beyond end of record\n");
      exit (2);
    }
  result = rec_buf[rec_idx++];
  result |= rec_buf[rec_idx++] << 8;
  return (result);
}


static word get_byte (void)
{
  if (rec_idx + 1 > rec_len)
    {
      fprintf (stderr, "\nByte beyond end of record\n");
      exit (2);
    }
  return (rec_buf[rec_idx++]);
}


static dword get_word_or_dword (void)
{
  return (rec_type & REC32 ? get_dword () : get_word ());
}


static void show_string (const byte *p)
{
  printf ("\"");
  while (*p != 0)
    {
      if (*p == '\\')
        printf ("\\\\");
      else if (*p < 0x20 || *p > 0x7f)
        printf ("\\%.3o", *p);
      else
        putchar (*p);
      ++p;
    }
  printf ("\"");
}


static void show_name (int index)
{
  if (index < 1 || index > lname_count)
    printf ("#%d", index);
  else
    show_string (lname_list[index-1]);
}


static void show_seg (int index)
{
  if (index < 1 || index > segment_count)
    printf ("#%d", index);
  else
    show_name (segment_list[index-1].name);
}


static void show_group (int index)
{
  if (index < 1 || index > group_count)
    printf ("#%d", index);
  else
    show_name (group_list[index-1].name);
}


static void show_ext (int index)
{
  if (index < 1 || index > ext_count)
    printf ("#%d", index);
  else
    show_string (ext_list[index-1].name);
}


static void dump (void)
{
  int x;

  x = 0;
  while (rec_idx < rec_len)
    {
      if (x >= 16)
        {
          printf ("\n");
          x = 0;
        }
      if (x == 0)
        printf (" ");
      printf (" %.2x", rec_buf[rec_idx++]);
      ++x;
    }
  if (x != 0)
    printf ("\n");
}


static void list_theadr (void)
{
  byte string[256];

  show_record ("THEADR");
  printf (" ");
  get_string (string);
  show_string (string);
  printf ("\n");
}


static void list_coment (void)
{
  byte string[256];
  int len;

  show_record ("COMENT");
  printf (": ");
  switch (rec_buf[1])
    {
    case 0x9f:
      rec_idx += 2;
      len = rec_len - rec_idx;
      memcpy (string, rec_buf + rec_idx, len);
      string[len] = 0;
      printf ("Default library: ");
      show_string (string);
      break;
    case 0xa1:
      if (rec_buf[3] == 'C' && rec_buf[4] == 'V')
        {
          debug_info = debug_cv;
          printf ("CodeView style (%d) debug tables", rec_buf[2]);
        }
      else if (rec_buf[3] == 'H' && rec_buf[4] == 'L')
        {
          debug_info = debug_hll;
          printf ("HLL style (%d) debug tables", rec_buf[2]);
        }
      else
        printf ("New OMF extension");
      break;
    case 0xa2:
      if (rec_buf[2] == 1)
        printf ("Link pass separator");
      else
        printf ("Unknown link pass: %.2x", rec_buf[2]);
      break;
    default:
      printf ("unknown comment class: %.2x", rec_buf[1]);
      break;
    }
  printf ("\n");
}


static void list_lnames (void)
{
  byte string[256];

  show_record ("LNAMES");
  printf (":\n");
  while (rec_idx < rec_len)
    {
      get_string (string);
      if (lname_count >= lname_alloc)
        {
          lname_alloc += 16;
          lname_list = xrealloc (lname_list,
                                 lname_alloc * sizeof (*lname_list));
        }
      lname_list[lname_count] = xstrdup (string);
      ++lname_count;
      printf ("  #%d: ", lname_count);
      show_string (string);
      printf ("\n");
    }
}


static void list_segdef (void)
{
  int attributes, alignment, combination, frame, offset, name, class, overlay;
  dword length;
  struct segment *p;

  frame = 0;                    /* Keep the compiler happy */

  show_record ("SEGDEF");
  printf (" #%d ", segment_count + 1);

  attributes = get_byte ();
  alignment = attributes >> 5;
  combination = (attributes >> 2) & 7;
  if (alignment == 0)
    {
      frame = get_word ();
      offset = get_byte ();
    }
  length = get_word_or_dword ();
  name = get_index ();
  class = get_index ();
  overlay = get_index ();
  if (!(rec_type & REC32) && (attributes & 2))
    length = 1L << 16;

  if (segment_count >= segment_alloc)
    {
      segment_alloc += 8;
      segment_list = xrealloc (segment_list,
                               segment_alloc * sizeof (*segment_list));
    }
  p = &segment_list[segment_count++];
  p->name = name;
  p->class = class;
  p->length = length;

  show_name (name);
  printf ("  ");
  switch (alignment)
    {
    case 0:
      printf ("AT %#.4x", frame);
      break;
    case 1:
      printf ("BYTE");
      break;
    case 2:
      printf ("WORD");
      break;
    case 3:
      printf ("PARA");
      break;
    case 4:
      printf ("PAGE");
      break;
    case 5:
      printf ("DWORD");
      break;
    case 6:
      printf ("unknown alignment: %d", alignment);
      break;
    }
  printf (" ");
  switch (combination)
    {
    case 0:
      printf ("PRIVATE");
      break;
    case 2:
    case 4:
    case 7:
      printf ("PUBLIC");
      break;
    case 5:
      printf ("STACK");
      break;
    case 6:
      printf ("COMMON");
      break;
    default:
      printf ("unknown combination: %d", combination);
      break;
    }
  if (attributes & 1)
    printf (" USE32");
  printf (" Length: %#lx", length);
  printf (" CLASS ");
  show_name (class);
  printf ("\n");
}


static void list_grpdef (void)
{
  int name, type, seg;

  show_record ("GRPDEF");
  printf (" #%d ", group_count + 1);

  name = get_index ();
  if (group_count >= group_alloc)
    {
      group_alloc += 8;
      group_list = xrealloc (group_list,
                             group_alloc * sizeof (*group_list));
    }
  group_list[group_count++].name = name;
  show_name (name);
  printf (":");
  while (rec_idx < rec_len)
    {
      type = get_byte ();
      if (type != 0xff)
        printf (" [unknown type: %#x] ", type);
      seg = get_index ();
      printf (" ");
      show_seg (seg);
    }
  printf ("\n");
}


static void list_extdef (void)
{
  int type;
  byte name[256];

  show_record ("EXTDEF");
  printf (":\n");

  while (rec_idx < rec_len)
    {
      printf ("  #%d: ", ext_count + 1);
      get_string (name);
      show_string (name);
      type = get_index ();
      printf (", type: %d\n", type);
      if (ext_count >= ext_alloc)
        {
          ext_alloc += 16;
          ext_list = xrealloc (ext_list,
                               ext_alloc * sizeof (*ext_list));
        }
      ext_list[ext_count++].name = xstrdup (name);
    }
}


static void list_pubdef (void)
{
  int type, group, seg;
  dword offset;
  byte name[256];

  show_record ("PUBDEF");
  group = get_index ();
  seg = get_index ();
  printf (" base group: ");
  show_group (group);
  printf (", base seg: ");
  show_seg (seg);
  if (seg == 0)
    printf (", frame=0x%.4x", get_word ());
  printf ("\n");

  while (rec_idx < rec_len)
    {
      printf ("  #%d: ", pub_count + 1);
      get_string (name);
      show_string (name);
      offset = get_word_or_dword ();
      type = get_index ();
      printf (", offset: %#lx, type: %d\n", offset, type);
      ++pub_count;
    }
}


static void list_symbols (void)
{
  int len, next, total_rec_len;
  dword offset, seg, length, reg;
  dword type, prologue, body, reserved, near_far, compiler_id;
  byte name[256];
  struct timestamp ts;

  total_rec_len = rec_len;
  while (rec_idx < rec_len)
    {
      len = get_byte () - 1;
      type = get_byte ();
      next = rec_idx + len;
      rec_len = next;
      printf ("  ");
      switch (type)
        {
        case 0x00:
          printf ("Begin ");
          offset = get_dword ();
          length = get_dword ();
          printf ("offset: %#lx, length: %#lx", offset, length);
          if (rec_idx < rec_len)
            {
              get_string (name);
              printf (" ");
              show_string (name);
            }
          break;
        case 0x01:
        case 0x0f:
          printf (type == 0x01 ? "Proc " : "Entry ");
          offset = get_dword ();
          type = get_word ();
          length = get_dword ();
          prologue = get_word ();
          body = get_dword ();
          reserved = get_word ();
          near_far = get_byte ();
          get_string (name);
          show_string (name);
          printf (", offset: %#lx, type: #%d\n", offset, type);
          printf ("       length: %#lx, pro: %#lx, pro+body: %#lx, "
                  "res: %#lx, nf: %#x",
                  length, prologue, body, reserved, near_far);
          break;
        case 0x02:
          printf ("End");
          break;
        case 0x04:
          printf ("Auto ");
          offset = get_dword ();
          type = get_word ();
          get_string (name);
          show_string (name);
          printf (", offset: %#lx, type: #%d", offset, type);
          break;
        case 0x05:
          printf ("Static ");
          offset = get_dword ();
          seg = get_word ();
          type = get_word ();
          get_string (name);
          show_string (name);
          printf (", offset: %#lx, seg: %d, type: #%d", offset, seg, type);
          break;
        case 0x0b:
          printf ("CodeLabel ");
          offset = get_dword ();
          near_far = get_byte ();
          get_string (name);
          printf ("offset: %#lx ", offset);
          show_string (name);
          break;
        case 0x0c:
          printf ("With");
          break;
        case 0x0d:
          printf ("Reg ");
          type = get_word ();
          reg = get_byte ();
          get_string (name);
          show_string (name);
          printf (", ");
          switch (reg)
            {
            case 0x00:
              printf ("AL");
              break;
            case 0x01:
              printf ("CL");
              break;
            case 0x02:
              printf ("DL");
              break;
            case 0x03:
              printf ("BL");
              break;
            case 0x04:
              printf ("AH");
              break;
            case 0x05:
              printf ("CH");
              break;
            case 0x06:
              printf ("DH");
              break;
            case 0x07:
              printf ("BH");
              break;
            case 0x08:
              printf ("AX");
              break;
            case 0x09:
              printf ("CX");
              break;
            case 0x0a:
              printf ("DX");
              break;
            case 0x0b:
              printf ("BX");
              break;
            case 0x0c:
              printf ("SP");
              break;
            case 0x0d:
              printf ("BP");
              break;
            case 0x0e:
              printf ("SI");
              break;
            case 0x0f:
              printf ("DI");
              break;
            case 0x10:
              printf ("EAX");
              break;
            case 0x11:
              printf ("ECX");
              break;
            case 0x12:
              printf ("EDX");
              break;
            case 0x13:
              printf ("EBX");
              break;
            case 0x14:
              printf ("ESP");
              break;
            case 0x15:
              printf ("EBP");
              break;
            case 0x16:
              printf ("ESI");
              break;
            case 0x17:
              printf ("EDI");
              break;
            case 0x18:
              printf ("ES"); /* error in documentation */
              break;
            case 0x19:
              printf ("CS");
              break;
            case 0x1a:
              printf ("SS");
              break;
            case 0x1b:
              printf ("DS");
              break;
            case 0x1c:
              printf ("FS");
              break;
            case 0x1d:
              printf ("GS");
              break;
            case 0x20:
              printf ("DX:AX");
              break;
            case 0x21:
              printf ("ES:BX");
              break;
            case 0x22:
              printf ("IP");
              break;
            case 0x23:
              printf ("FLAGS");
              break;
            case 0x24:
              printf ("EFLAGS");
              break;
            case 0x80:
            case 0x81:
            case 0x82:
            case 0x83:
            case 0x84:
            case 0x85:
            case 0x86:
            case 0x87:
              printf ("ST(%d)", reg - 0x80);
              break;
            default:
              printf ("unknown register: %#.2x", reg);
            }
          break;
        case 0x0e:
          printf ("Constant");
          break;
        case 0x10:
          printf ("Skip");
          break;
        case 0x11:
          printf ("ChangSeg ");
          reserved = get_dword ();
          printf ("%#lx", reserved);
          break;
        case 0x12:
          printf ("Typedef");
          break;
        case 0x13:
          printf ("Public");
          break;
        case 0x14:
          printf ("Member");
          break;
        case 0x15:
          printf ("Based");
          break;
        case 0x16:
          printf ("Tag ");
          type = get_word ();
          printf ("type: #%d, name: ", type);
          get_string (name);
          show_string (name);
          break;
        case 0x17:
          printf ("Table");
          break;
        case 0x18:
          printf ("Map");
          break;
        case 0x19:
          printf ("Class ");
          type = get_word ();
          printf ("type: #%d, name: ", type);
          get_string (name);
          show_string (name);
          break;
        case 0x40:
          printf ("CuInfo: ");
          compiler_id = get_byte ();
          switch (compiler_id)
            {
            case 0x01:
              printf ("C");
              break;
            case 0x02:
              printf ("C++");
              break;
            default:
              printf ("unknown compiler id: %#x", compiler_id);
              break;
            }
          printf (", options: ");
          get_string (name);
          show_string (name);
          printf (", time stamp:");
          get_mem (&ts, sizeof (ts));
          printf ("%d/%.2d/%.2d %.2d:%.2d:%.2d.%.2d",
                  ts.year, ts.month, ts.day, ts.hours,
                  ts.minutes, ts.seconds, ts.hundredths);
          break;
        default:
          printf ("unknown SST type %#.2x", type);
          break;
        }
      printf ("\n");
      rec_idx = next;
      rec_len = total_rec_len;
    }

}



static void show_fid (void)
{
  int fid, index, i;
  byte str[256];
  dword d;

  fid = get_byte ();
  switch (fid)
    {
    case 0x80:
      printf ("(nil)");
      break;
    case 0x81:
      printf ("void");
      break;
    case 0x82:
      get_string (str);
      show_string (str);
      break;
    case 0x83:
      index = get_word ();
      switch (index)
        {
        case 0x80:
          printf ("signed char");
          break;
        case 0x81:
          printf ("signed short");
          break;
        case 0x82:
          printf ("signed long");
          break;
        case 0x84:
          printf ("unsigned char");
          break;
        case 0x85:
          printf ("unsigned short");
          break;
        case 0x86:
          printf ("unsigned long");
          break;
        case 0x88:
          printf ("float");
          break;
        case 0x89:
          printf ("double");
          break;
        case 0x8a:
          printf ("long double");
          break;
        case 0x97:
          printf ("void");
          break;
        case 0xa0:
          printf ("signed char *");
          break;
        case 0xa1:
          printf ("signed short *");
          break;
        case 0xa2:
          printf ("signed long *");
          break;
        case 0xa4:
          printf ("signed char *");
          break;
        case 0xa5:
          printf ("signed short *");
          break;
        case 0xa6:
          printf ("signed long *");
          break;
        case 0xa8:
          printf ("float *");
          break;
        case 0xa9:
          printf ("double *");
          break;
        case 0xaa:
          printf ("long double *");
          break;
        case 0xb7:
          printf ("void *");
          break;
        default:
          printf ("#%d", index);
          break;
        }
      break;
    case 0x85:
      i = get_word ();
      printf ("%u", i);
      break;
    case 0x86:
      d = get_dword ();
      printf ("%lu", d);
      break;
    case 0x88:
      i = get_byte ();
      if (i >= 0x80)
        i -= 0x80;
      printf ("%d", i);
      break;
    case 0x89:
      i = get_word ();
      if (i >= 0x8000)
        i -= 0x8000;
      printf ("%d", i);
      break;
    case 0x8a:
      d = get_dword ();
      printf ("%ld", d);
      break;
    case 0x8b:
      i = get_byte ();
      printf ("%u", i);
      break;
    default:
      fprintf (stderr, "unknown FID: %#.2x\n", fid);
      exit (2);
    }
}


static void list_vis (void)
{
  int flags;

  flags = get_byte ();
  switch (flags)
    {
    case 0x00:
      printf ("private");
      break;
    case 0x01:
      printf ("protected ");
      break;
    case 0x02:
      printf ("public");
      break;
    default:
      printf ("unknown visibility: %#x", flags);
      break;
    }
}


static void list_types (void)
{
  int i, total_rec_len, general, type, qual, len, next, flags;
  int args, max_args, fields;
  dword size;

  total_rec_len = rec_len;
  while (rec_idx < rec_len)
    {
      general = get_byte ();
      len = get_word ();
      next = rec_idx + len;
      type = get_byte ();
      qual = get_byte ();
      printf ("  #%d: ", type_index++);
      if (next < total_rec_len)
        rec_len = next;
      if (general != 0x01)
        {
          printf ("Unknown sub-record type: %#.2x\n", general);
          dump ();
        }
      else
        {
          switch (type)
            {
            case 0x40:
              printf ("Class: ");
              if (qual != 0)
                printf ("(qual: %#x) ", qual);
              size = get_dword ();
              fields = get_word ();
              printf ("size: %#lx, %d fields: ", size, fields);
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x41:
              printf ("Subclass: ");
              if (qual != 0)
                printf ("(qual: %#x) ", qual);
              list_vis ();
              printf (", class: ");
              show_fid ();
              printf (", whatever: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x42:
              printf ("Friend function: ");
              if (qual != 0)
                printf ("(qual: %#x) ", qual);
              printf ("type: ");
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x45:
              printf ("Method: ");
              if (qual != 0)
                printf ("(qual: %#x) ", qual);
              list_vis ();
              printf (", whatever:");
              for (i = 0; i < 5; ++i)
                printf (" %.2x", get_byte ());
              flags = get_byte ();
              if (flags & 0x01)
                printf (" static");
              if (flags & 0x02)
                printf (" inline");
              if (flags & 0x10)
                printf (" virtual");
              if (flags & ~0x13)
                printf (" unknown flags: %#x", flags);
              printf ("\n");
              args = get_word ();
              max_args = get_word ();
              printf ("    %d argument%s (%d maximum), returns ",
                      args, (args == 1 ? "" : "s"), max_args);
              show_fid ();
              printf (", argument list: ");
              show_fid ();
              printf (",\n    name: ");
              show_fid ();
              printf (", whatever: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x46:
              printf ("Class field: ");
              if (qual & 1)
                printf ("static ");
              if ((qual & ~1) != 0)
                printf ("(qual: %#x) ", qual);
              list_vis ();
              printf (", type: ");
              show_fid ();
              printf (", offset: ");
              show_fid ();
              printf (",\n    whatever: ");
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x48:
              printf ("Reference: ");
              if (qual != 0)
                printf ("(qual: %#x) ", qual);
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x52:
              printf ("Set: ");
              if (qual != 0)
                printf ("qual: %#x, ", qual);
              printf ("type: ");
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x54:
              printf ("Function: ");
              args = get_word ();
              max_args = get_word ();
              printf ("qual: %#x, %d argument%s (%d maximum), returns ",
                      qual, args, (args == 1 ? "" : "s"), max_args);
              show_fid ();
              printf (",\n    argument list: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x57:
              printf ("Stack: ");
              if (qual & 0x01)
                printf ("32-bit ");
              if (qual & 0x02)
                printf ("far ");
              if (qual & ~0x03)
                printf ("qual: %#x, ", qual);
              size = get_dword ();
              printf ("size: %#x, name: ", size);
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x5c:
              printf ("Bit string: ");
              if (qual & 1)
                printf ("varying, ");
              if (qual & 2)
                printf ("signed, ");
              if (qual & 4)
                printf ("word alignment, ");
              if (qual & 8)
                printf ("display as value, ");
              if (qual & 0x10)
                printf ("descriptor, ");
              if (qual & ~0x1f)
                printf ("qual: %#x, ", qual);
              printf ("offset: %d, size: ", get_byte ());
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x5d:
              printf ("User defined type: ");
              if (qual != 0)
                printf ("qual: %#x, ", qual);
              printf ("type: ");
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x6f:
              printf ("Subrange: ");
              if (qual != 0)
                printf ("qual: %#x, ", qual);
              printf ("type: ");
              show_fid ();
              printf (", start: ");
              show_fid ();
              printf (", end: ");
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x78:
              printf ("Array: ");
              if (qual & 1)
                printf ("col-maj ");
              else
                printf ("row-maj ");
              if (qual & 2)
                printf ("(packed) ");
              if (qual & 4)
                printf ("desc provided) ");
              if (qual & ~7)
                printf ("(qual: %#x) ", qual);
              size = get_dword ();
              printf ("size: %#lx, bounds type: ", size);
              show_fid ();
              printf (", elements type: ");
              show_fid ();
              printf (", name: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x79:
              printf ("Structure: ");
              if (qual & 1)
                printf ("(packed) ");
              if (qual & ~1)
                printf ("(qual: %#x) ", qual);
              size = get_dword ();
              fields = get_word ();
              printf ("size: %#lx, %d fields: ", size, fields);
              show_fid ();
              printf (", names: ");
              show_fid ();
              printf (", tag: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x7a:
              printf ("Pointer: ");
              show_fid ();
              if (rec_idx < rec_len)
                {
                  printf (", name: ");
                  show_fid ();
                }
              printf ("\n");
              dump ();
              break;
            case 0x7b:
              printf ("Enum: ");
              if (qual != 0)
                printf ("(qual: %#x) ", qual);
              printf ("type: ");
              show_fid ();
              printf (", values: ");
              show_fid ();
              printf (", minimum: ");
              show_fid ();
              printf (", maximum: ");
              show_fid ();
              printf ("\n");
              dump ();
              break;
            case 0x7f:
              printf ("List");
              switch (qual)
                {
                case 0x01:
                  printf (" (field types)");
                  break;
                case 0x02:
                  printf (" (field names and offsets)");
                  break;
                case 0x03:
                  printf (" (enumeration)");
                  break;
                case 0x04:
                  printf (" (arguments)");
                  break;
                }
              printf (":\n");
              while (rec_idx < rec_len)
                {
                  printf ("    ");
                  if (qual == 0x04)
                    {
                      flags = get_byte ();
                      printf ("flags: %#x ", flags);
                    }
                  show_fid ();
                  switch (qual)
                    {
                    case 0x02:
                      printf (", offset: ");
                      show_fid ();
                      break;
                    case 0x03:
                      printf (", value: ");
                      show_fid ();
                      break;
                    }
                  printf ("\n");
                }
              dump ();
              break;
            default:
              printf ("unknown complex type: %#.2x\n", type);
              dump ();
              break;
            }
        }
      rec_idx = next;
      rec_len = total_rec_len;
    }
}


static void list_ledata (void)
{
  int len;
  dword offset, seg;
  char *name;

  show_record ("LEDATA");

  seg = get_index ();
  offset = get_word_or_dword ();
  len = rec_len - rec_idx;

  printf (" ");
  show_seg (seg);
  printf (" offset: %#lx length: %#lx\n", offset, len);
  name = lname_list[segment_list[seg-1].name-1];
  if (strcmp (name, "$$SYMBOLS") == 0 && debug_info == debug_hll)
    list_symbols ();
  else if (strcmp (name, "$$TYPES") == 0 && debug_info == debug_hll)
    list_types ();
  else
    dump ();
}


static void list_target (int method)
{
  switch (method)
    {
    case 0:
      printf ("T0 (seg ");
      show_seg (get_index ());
      printf (")");
      break;
    case 1:
      printf ("T1 (group ");
      show_group (get_index ());
      printf (")");
      break;
    case 2:
      printf ("T2 (ext ");
      show_ext (get_index ());
      printf (")");
      break;
    default:
      printf ("T%d (index #%d)", method & 3, get_index ());
                break;
    }
}


static void list_frame (int method)
{
  switch (method)
    {
    case 0:
      printf ("F0 (seg ");
      show_seg (get_index ());
      printf (")");
      break;
    case 1:
      printf ("F1 (group ");
      show_group (get_index ());
      printf (")");
      break;
    case 2:
      printf ("F2 (ext");
      show_ext (get_index ());
      printf (")");
      break;
    case 4:
      printf ("F4 (cur seg)");
      break;
    case 5:
      printf ("F5 (target)");
      break;
    default:
      printf ("F%d", method);
      if (method <= 2)
        printf (" (index #%d)", get_index ());
      break;
    }
}


static void list_fixupp (void)
{
  int first, locat, method, offset, fix_data;

  show_record ("FIXUPP");
  printf (":\n");
  while (rec_idx < rec_len)
    {
      first = get_byte ();
      if (first & 0x80)
        {
          printf ("  FIXUP ");
          if (first & 0x40)
            printf ("seg-rel, ");
          else
            printf ("self-rel, ");
          locat = (first >> 2) & 0x0f;
          switch (locat)
            {
            case 0:
              printf ("LOW-8");
              break;
            case 1:
              printf ("OFFSET-16");
              break;
            case 2:
              printf ("SEL-16");
              break;
            case 3:
              printf ("FAR-16:16");
              break;
            case 4:
              printf ("HIGH-8");
              break;
            case 5:
              printf ("OFFSET-16(LR)");
              break;
            case 9:
              printf ("OFFSET-32");
              break;
            case 11:
              printf ("FAR-16:32");
              break;
            case 13:
              printf ("OFFSET-32(LR)");
              break;
            default:
              printf ("unknown location: %d", locat);
              break;
            }
          offset = get_byte ();
          offset |= (first & 3) << 8;
          printf (", offset: %#x, ", offset);
          fix_data = get_byte ();
          if (fix_data & 0x80)
            printf ("frame thread %d", (fix_data >> 4) & 3);
          else
            list_frame ((fix_data >> 4) & 7);
          if (fix_data & 0x08)
            printf (", target thread %d P=%d",
                    fix_data & 3, (fix_data >> 2) & 1);
          else
            {
              printf (", ");
              list_target (fix_data & 3);
            }
          if (!(fix_data & 0x04))
            printf (", disp: %#lx", get_word_or_dword ());
        }
      else
        {
          method = (first >> 2) & 7;
          printf ("  THREAD %d ", first & 3);
          if (first & 0x40)
            {
              list_frame (method);
              printf (", P=%d", (first >> 4) & 1);
            }
          else
            list_target (method & 3);
        }
      printf ("\n");
    }
}


static void list_hll_linnum (void)
{
  int group, segment;
  dword uk1, uk2, uk3, i, addr, files;
  word line;
  byte index, flags;
  byte name[256];

  group = get_index ();
  segment = get_index ();
  printf (": group: %d, segment: ", group);
  show_seg (segment);
  if (!linnum_started)
    {
      uk1 = get_dword ();
      line_count = get_dword ();
      names_length = get_dword ();
      printf (", uk1: %#lx, %ld line%s, name table length: %#lx",
              uk1, line_count, (line_count == 1 ? "" : "s"), names_length);
      linnum_started = TRUE;
    }
  printf ("\n");
  while (rec_idx < rec_len && line_count != 0)
    {
      line = get_word ();
      index = get_byte ();
      flags = get_byte ();
      addr = get_dword ();
      printf ("  Line %6d of file %d at 0x%.8lx", line, index, addr);
      if (flags & 0x80)
        printf (", #include");
      if (flags & 0x7f)
        printf (", flags: %#x", flags);
      printf ("\n");
      --line_count;
    }

  /* Note: this may fail if the name table is split */

  if (rec_idx < rec_len)
    {
      uk2 = get_dword ();
      uk3 = get_dword ();
      files = get_dword ();
      printf ("  uk2: %#lx, uk3: %#lx, number of files: %d\n",
              uk2, uk3, files);
      for (i = 0; i < files; ++i)
        {
          printf ("  #%d: ", i + 1);
          get_string (name);
          show_string (name);
          printf ("\n");
        }
    }
}


static void list_modend (void)
{
  int type, end_data;

  show_record ("MODEND");
  printf (": ");
  type = get_byte ();
  if (type & 0x80)
    printf ("main ");
  if (type & 0x40)
    {
      if (type & 0x01)
        printf ("rel-");
      printf ("start: ");
      end_data = get_byte ();
      if (end_data & 0x80)
        printf ("frame thread %d", (end_data >> 4) & 3);
      else
        list_frame ((end_data >> 4) & 7);
      if (end_data & 0x08)
        printf (", target thread %d P=%d",
                end_data & 3, (end_data >> 2) & 1);
      else
        {
          printf (", ");
          list_target (end_data & 3);
        }
      if (!(end_data & 0x04))   /* This bit must always be clear! */
        printf (", disp: %#lx", get_word_or_dword ());
    }
  printf ("\n");
  dump ();
}


static void list_omf (const char *fname)
{
  FILE *f;
  struct omf_rec rec;

  f = fopen (fname, "rb");
  if (f == NULL)
    {
      perror (fname);
      exit (2);
    }
  lname_count = 0; lname_alloc = 0; lname_list = NULL;
  segment_count = 0; segment_alloc = 0; segment_list = NULL;
  group_count = 0; group_alloc = 0; group_list = NULL;
  ext_count = 0; ext_alloc = 0; ext_list = NULL;
  pub_count = 0;
  type_index = 512;
  debug_info = debug_default; linnum_started = FALSE;
  do
    {
      if (fread (&rec, sizeof (rec), 1, f) != 1)
        goto failure;
      rec_type = rec.rec_type;
      rec_len = rec.rec_len;
      rec_idx = 0;
      if (rec_len > sizeof (rec_buf))
        {
          fprintf (stderr, "%s: Record too long", fname);
          exit (2);
        }
      if (fread (rec_buf, rec_len, 1, f) != 1)
        goto failure;
      /*...check checksum...*/
      --rec_len;
      switch (rec_type)
        {
        case THEADR:
          list_theadr ();
          break;
        case COMENT:
          list_coment ();
          break;
        case MODEND:
        case MODEND|REC32:
          list_modend ();
          break;
        case EXTDEF:
          list_extdef ();
          break;
        case TYPDEF:
          show_record ("TYPDEF");
          printf ("\n");
          dump ();
          break;
        case PUBDEF:
        case PUBDEF|REC32:
          list_pubdef ();
          break;
        case LINNUM:
        case LINNUM|REC32:
          show_record ("LINNUM");
          if (debug_info == debug_hll)
            list_hll_linnum ();
          else
            {
              printf ("\n");
              dump ();
            }
          break;
        case LNAMES:
          list_lnames ();
          break;
        case SEGDEF:
        case SEGDEF|REC32:
          list_segdef ();
          break;
        case GRPDEF:
          list_grpdef ();
          break;
          break;
        case FIXUPP:
        case FIXUPP|REC32:
          list_fixupp ();
          break;
        case LEDATA:
        case LEDATA|REC32:
          list_ledata ();
          break;
        case LIDATA:
          show_record ("LIDATA");
          printf ("\n");
          dump ();
          break;
        case COMDEF:
          show_record ("COMDEF");
          printf ("\n");
          dump ();
          break;
        default:
          printf ("Unknown record type: %.2x\n", rec.rec_type);
          dump ();
          break;
        }
    } while (rec_type != MODEND && rec_type != (MODEND|REC32));
  fclose (f);
  if (lname_list != NULL)
    free (lname_list);
  return;

failure:
  if (ferror (f))
    perror (fname);
  else
    fprintf (stderr, "%s: Unexpected end of file", fname);
  exit (2);
}


int main (int argc, char *argv[])
{
  if (argc != 2)
    {
      puts ("Usage: listomf <input_file>.obj");
      return (1);
    }
  list_omf (argv[1]);
  return (0);
}
